import boto3
import json
import logging
import os

from boto3utils import s3
from boto3.dynamodb.conditions import Key
from datetime import datetime, timedelta, timezone
from typing import Dict, Optional, List

# envvars
CATALOG_BUCKET = os.getenv('CIRRUS_CATALOG_BUCKET')

STATES = ['PROCESSING', 'COMPLETED', 'FAILED', 'INVALID']

# logging
logger = logging.getLogger(__name__)


class StateDB:

    def __init__(self, table_name: str=os.getenv('CIRRUS_STATE_DB', 'test')):
        """Initialize a StateDB instance using the Cirrus State DB table

        Args:
            table_name (str, optional): The Cirrus StateDB Table name. Defaults to os.getenv('CIRRUS_STATE_DB', None).
        """
        # initialize client
        self.db = boto3.resource('dynamodb')
        self.table_name = table_name
        self.table = self.db.Table(table_name)

    def delete_item(self, catid: str):
        key = self.catid_to_key(catid)
        response = self.table.delete_item(Key=key)
        logger.debug("Removed item", extra=key)
        return response

    def delete(self):
        # delete table (used for testing)
        self.table.delete()
        self.table.wait_until_not_exists()        

    def get_dbitem(self, catid: str) -> Dict:
        """Get a DynamoDB item

        Args:
            catid (str): Catalog ID

        Raises:
            Exception: Error getting item

        Returns:
            Dict: DynamoDB Item
        """
        key=self.catid_to_key(catid)
        try:
            response = self.table.get_item(Key=key)
            return response.get('Item', None)
        except Exception as err:
            msg = "Error fetching item"
            logger.error(msg, extra=key.update({'error': err}), exc_info=True)
            raise Exception(msg)

    def get_dbitems(self, catids: List[str]) -> List[Dict]:
        """Get multiple DynamoDB Items

        Args:
            catids (List[str]): A List of Catalog IDs

        Raises:
            Exception: Error getting items

        Returns:
            List[Dict]: A list of DynamoDB Items
        """
        try:
            resp = self.db.meta.client.batch_get_item(RequestItems={
                self.table_name: {
                    'Keys': [self.catid_to_key(id) for id in catids]
                }
            })
            items = []
            for r in resp['Responses'][self.table_name]:
                items.append(r)
            logger.debug(f"Fetched {len(items)} items")
            return items
        except Exception as err:
            msg = f"Error fetching items"
            logger.error(msg, exc_info=True)
            raise Exception(msg)

    def get_counts(self, collections_workflow: str, state: str=None, since: str=None, limit: int=None) -> Dict:
        """Get counts by query

        Args:
            collections_workflow (str): /-separated list of collections (input or output depending on index)
            state (Optional[str], optional): State of Items to get. Defaults to None.
            since (Optional[str], optional): Get Items since this amount of time in the past. Defaults to None.
            limit (int, optional): The max number to return, anything over will be reported as "<limit>+", e.g. "1000+"

        Returns:
            Dict: JSON containing counts key with counts for each state requested
        """
        counts = 0
        resp = self.query(collections_workflow, state=state, since=since, select='COUNT')
        counts = resp['Count']
        while 'LastEvaluatedKey' in resp:
            resp = self.query(collections_workflow, state=state, since=since, select='COUNT',
                              ExclusiveStartKey=resp['LastEvaluatedKey'])
            counts += resp['Count']
            if limit and counts > limit:
                counts = f"{limit}+"
                break

        return counts

    def get_items_page(self, collections_workflow: str,
                       state: Optional[str]=None, since: Optional[str]=None,
                       limit=100, nextkey: str=None, sort_ascending: Optional[bool]=False,
                       sort_index: Optional[str]=None) -> List[Dict]:
        """Get Items by query

        Args:
            collections_workflow (str): /-separated list of input collections_workflow
            state (Optional[str], optional): State of Items to get (PROCESSING, COMPLETED, FAILED, INVALID)
            since (Optional[str], optional): Get Items since this amount of time in the past. Defaults to None.
            sort_ascending (Optional[bool], optional): Determines which direction the index of the results will be sorted. Defaults to False.
            sort_index (Optional[str], optional): Determines which index to use for sorting, if not applying a filter (state_updated, updated). Defaults to None.

        Returns:
            Dict: List of Items
        """
        items = {
            'items': []
        }
        if nextkey:
            dbitem = self.get_dbitem(nextkey)
            startkey = { key: dbitem[key] for key in ['collections_workflow', 'itemids', 'state_updated', 'updated']}
            resp = self.query(collections_workflow, state=state, since=since, sort_ascending=sort_ascending, sort_index=sort_index, Limit=limit, ExclusiveStartKey=startkey, )
        else:
            resp = self.query(collections_workflow, state=state, since=since,  sort_ascending=sort_ascending, sort_index=sort_index, Limit=limit)
        for i in resp['Items']:
            items['items'].append(self.dbitem_to_item(i))
        if 'LastEvaluatedKey' in resp:
            items['nextkey'] = self.key_to_catid(resp['LastEvaluatedKey'])
        return items

    def get_items(self, *args, limit=None, **kwargs) -> Dict:
        """Get items from database

        Args:
            limit (int, optional): Maximum number of items to return. Defaults to None.

        Returns:
            Dict: StateDB Items
        """
        resp = self.get_items_page(*args, **kwargs)
        items = resp['items']
        while 'nextkey' in resp and (limit is None or len(items) < limit):
            resp = self.get_items_page(*args, nextkey=resp['nextkey'], **kwargs)
            items += resp['items']
        if limit is None or len(items) < limit:
            return items
        return items[:limit]

    def get_state(self, catid: str) -> str:
        """Get current state of Item

        Args:
            catid (str): The catalog ID

        Returns:
            str: Current state: PROCESSING, COMPLETED, FAILED, INVALID
        """
        response = self.table.get_item(Key=self.catid_to_key(catid))
        if 'Item' in response:
            return response['Item']['state_updated'].split('_')[0]
        else:
            # assuming no such item in database
            return ""

    def get_states(self, catids: List[str]) -> Dict[str, str]:
        """Get current state of items

        Args:
            catids (List[str]): List of catalog IDs

        Returns:
            Dict[str, str]: Dictionary of catalog IDs to state
        """
        states = {}
        for dbitem in self.get_dbitems(catids):
            item = self.dbitem_to_item(dbitem)
            states[item['catid']] = item['state']
        return states

    def claim_processing(self, catid):
        """ Sets catid to PROCESSING to claim it (preventing other runs) """
        now = datetime.now(timezone.utc).isoformat()
        key = self.catid_to_key(catid)

        expr = (
            'SET '
            'created = if_not_exists(created, :created), '
            'state_updated=:state_updated, updated=:updated'
        )
        response = self.table.update_item(
            Key=key,
            UpdateExpression=expr,
            ConditionExpression='NOT begins_with(state_updated, :proc)',
            ExpressionAttributeValues={
                ':created': now,
                ':state_updated': f"PROCESSING_{now}",
                ':updated': now,
                ':proc': "PROCESSING"
            }
        )
        logger.debug("Claimed processing", extra=key)
        return response

    def set_processing(self, catid, execution):
        """ Adds execution to existing item or creates new """
        now = datetime.now(timezone.utc).isoformat()
        key = self.catid_to_key(catid)

        expr = (
            'SET '
            'created = if_not_exists(created, :created), '
            'state_updated=:state_updated, updated=:updated, '
            'executions = list_append(if_not_exists(executions, :empty_list), :exes)'
        )
        response = self.table.update_item(
            Key=key,
            UpdateExpression=expr,
            ExpressionAttributeValues={
                ':created': now,
                ':state_updated': f"PROCESSING_{now}",
                ':updated': now,
                ':empty_list': [],
                ':exes': [execution]
            }
        )
        logger.debug("Add execution", extra=key.update({'execution': execution}))
        return response   

    def set_completed(self, catid: str, outputs: List[str]) -> str:
        """Set this catalog as COMPLETED

        Args:
            catid (str): The Cirrus Catalog
            outputs ([str]): List of URLs to output Items

        Returns:
            str: DynamoDB response
        """
        now = datetime.now(timezone.utc).isoformat()
        key = self.catid_to_key(catid)

        expr = (
            'SET '
            'created = if_not_exists(created, :created), '
            'state_updated=:state_updated, updated=:updated, '
            'outputs=:outputs'
        )
        response = self.table.update_item(
            Key=key,
            UpdateExpression=expr,
            ExpressionAttributeValues={
                ':created': now,
                ':state_updated': f"COMPLETED_{now}",
                ':updated': now,
                ':outputs': outputs
            }
        )
        logger.debug("set completed", extra=key.update({'outputs': outputs}))
        return response

    def set_failed(self, catid, msg):
        """ Adds new item as failed """
        """ Adds new item with state function execution """
        now = datetime.now(timezone.utc).isoformat()
        key = self.catid_to_key(catid)

        expr = (
            'SET '
            'created = if_not_exists(created, :created), '
            'state_updated=:state_updated, updated=:updated, '
            'last_error=:last_error'
        )
        response = self.table.update_item(
            Key=key,
            UpdateExpression=expr,
            ExpressionAttributeValues={
                ':created': now,
                ':state_updated': f"FAILED_{now}",
                ':updated': now,
                ':last_error': msg
            }
        )
        logger.debug("set failed", extra=key.update({'last_error': msg}))
        return response

    def set_invalid(self, catid: str, msg: str) -> str:
        """Set this catalog as INVALID

        Args:
            catid (str): The Cirrus Catalog
            msg (str): An error message to include in DynamoDB Item

        Returns:
            str: DynamoDB response
        """
        now = datetime.now(timezone.utc).isoformat()
        key = self.catid_to_key(catid)

        expr = (
            'SET '
            'created = if_not_exists(created, :created), '
            'state_updated=:state_updated, updated=:updated, '
            'last_error=:last_error'
        )
        response = self.table.update_item(
            Key=key,
            UpdateExpression=expr,
            ExpressionAttributeValues={
                ':created': now,
                ':state_updated': f"INVALID_{now}",
                ':updated': now,
                ':last_error': msg
            }
        )
        logger.debug("set invalid", extra=key.update({'last_error': msg}))
        return response

    def query(self, collections_workflow: str, state: str=None, since: str=None,
              select: str='ALL_ATTRIBUTES', sort_ascending: bool=False, sort_index: str='updated', **kwargs) -> Dict:
        """Perform a single Query on a DynamoDB index

        Args:
            collections_workflow (str): The complete has to query
            state (str, optional): The state of the Item. Defaults to None.
            since (str, optional): Query for items since this time. Defaults to None.
            select (str, optional): DynamoDB Select statement (ALL_ATTRIBUTES, COUNT). Defaults to 'ALL_ATTRIBUTES'.
            sort_ascending (bool, optional): Determines which direction the index of the results will be sorted.
                Defaults to False/Descending.
            sort_index (str, optional): Determines which index to use for sorting, if not applying a filter (default, state_updated, updated)
                If default, sorting will use primary index and sort by item_ids

        Returns:
            Dict: DynamoDB response
        """
        index = None if sort_index == 'default' else sort_index
        

        # always use the hash of the table which is same in all Global Secondary Indices
        expr = Key('collections_workflow').eq(collections_workflow)
        if since:
            start = datetime.now(timezone.utc) - self.since_to_timedelta(since)
            begin = f"{start.isoformat()}"
            end = f"{datetime.now(timezone.utc).isoformat()}"
            if state:
                index = 'state_updated'
                expr = expr & Key(index).between(f"{state}_{begin}", f"{state}_{end}")
            else:
                index = 'updated'
                expr = expr & Key(index).between(begin, end)
        elif state:
            index = 'state_updated'
            expr = expr & Key(index).begins_with(state)

        keys = ['collections_workflow', 'itemids']
        if index:
            keys.append(index)
        if 'ExclusiveStartKey' in kwargs:
            kwargs['ExclusiveStartKey'] = {k: kwargs['ExclusiveStartKey'][k] for k in keys}

        if index:
            resp = self.table.query(IndexName=index, KeyConditionExpression=expr, Select=select, ScanIndexForward=sort_ascending, **kwargs)
        else:
            resp = self.table.query(KeyConditionExpression=expr, Select=select,  ScanIndexForward=sort_ascending, **kwargs)

        return resp

    @classmethod
    def catid_to_key(cls, catid: str) -> Dict:
        """Create DynamoDB Key from catalog ID

        Args:
            catid (str): The catalog ID

        Returns:
            Dict: Dictionary containing the DynamoDB Key
        """
        parts1 = catid.split('/workflow-')
        parts2 = parts1[1].split('/', maxsplit=1)
        key = {
            'collections_workflow': parts1[0] + f"_{parts2[0]}",
            'itemids': '' if len(parts2) == 1 else parts2[1]
        }
        return key

    @classmethod
    def key_to_catid(cls, key: Dict) -> str:
        """Get catalog ID given a DynamoDB Key

        Args:
            key (Dict): DynamoDB Key

        Returns:
            str: Catalog ID
        """
        parts = key['collections_workflow'].rsplit('_', maxsplit=1)
        return f"{parts[0]}/workflow-{parts[1]}/{key['itemids']}"

    @classmethod
    def get_input_catalog_url(self, dbitem):
        catid = self.key_to_catid(dbitem)
        return f"s3://{CATALOG_BUCKET}/{catid}/input.json"

    @classmethod
    def dbitem_to_item(cls, dbitem: Dict, region: str=os.getenv('AWS_REGION', 'us-west-2')) -> Dict:
        state, updated = dbitem['state_updated'].split('_')
        collections, workflow = dbitem['collections_workflow'].rsplit('_', maxsplit=1)
        item = {
            "catid": cls.key_to_catid(dbitem),
            "collections": collections,
            "workflow": workflow,
            "items": dbitem['itemids'],
            "state": state,
            "created": dbitem['created'],
            "updated": dbitem['updated'],
            "catalog": cls.get_input_catalog_url(dbitem)
        }
        if 'executions' in dbitem:
            base_url = f"https://{region}.console.aws.amazon.com/states/home?region={region}#/executions/details/"
            item['executions'] = [base_url + f"{e}" for e in dbitem['executions']]
        if 'outputs' in dbitem:
            item['outputs'] = dbitem['outputs']
        if 'last_error' in dbitem:
            item['last_error'] = dbitem['last_error']
        return item

    @classmethod
    def since_to_timedelta(cls, since: str) -> timedelta:
        """Convert a `since` field to a timedelta.

        Args:
            since (str): Contains an integer followed by a unit letter: 'd' for days, 'h' for hours, 'm' for minutes

        Returns:
            timedelta: [description]
        """
        unit = since[-1]
        # days, hours, or minutes
        assert(unit in ['d', 'h', 'm'])
        days = int(since[0:-1]) if unit == 'd' else 0
        hours = int(since[0:-1]) if unit == 'h' else 0
        minutes = int(since[0:-1]) if unit == 'm' else 0
        return timedelta(days=days, hours=hours, minutes=minutes)



import logging
import logging.config
from os import getenv


config = {
  "version": 1,
  "disable_existing_loggers": False,
  "formatters": {
      "standard": {
          "format": "%(asctime)s %(name)s %(levelname)s %(message)s",
          "datefmt": "%Y-%m-%dT%H:%M:%S%z",
      },
      "json": {
          "format": "%(asctime)s %(name)s %(levelname)s %(message)s",
          "datefmt": "%Y-%m-%dT%H:%M:%S%z",
          "class": "pythonjsonlogger.jsonlogger.JsonFormatter"
      }
  },
  "handlers": {
      "standard": {
          "class": "logging.StreamHandler",
          "formatter": "json"
      }
  },
  "loggers": {
      "": {
          "propagate": False
      },
      "lambda_function": {
          "handlers": ["standard"],
          "level": getenv('CIRRUS_LOG_LEVEL', 'DEBUG')
      },
      "feeder": {
          "handlers": ["standard"],
          "level": getenv('CIRRUS_LOG_LEVEL', 'DEBUG')
      },
      "task": {
          "handlers": ["standard"],
          "level": getenv('CIRRUS_LOG_LEVEL', 'DEBUG')
      },
      "cirruslib": {
          "handlers": ["standard"],
          "level": getenv('CIRRUS_LOG_LEVEL', 'DEBUG')
      }
  }
}

logging.config.dictConfig(config)


class DynamicLoggerAdapter(logging.LoggerAdapter):

    def __init__(self, *args, keys=None, **kwargs):
        super(DynamicLoggerAdapter, self).__init__(*args, **kwargs)
        self.keys = keys

    def process(self, msg, kwargs):
        if self.keys is not None:
            kwargs = {k: self.extra[k] for k in self.keys if k in self.extra}
            return (msg, {"extra": kwargs})
        else:
            return (msg, kwargs)


def get_task_logger(*args, catalog, **kwargs):
    _logger = logging.getLogger(*args, **kwargs)
    logger = DynamicLoggerAdapter(_logger, catalog, keys=['id', 'stac_version'])
    return logger
    



from __future__ import annotations

import boto3
import json
import logging
import os
import re
import uuid
from datetime import datetime, timezone
from typing import Dict, Optional, List

from boto3utils import s3
from cirruslib.statedb import StateDB
from cirruslib.logging import get_task_logger
from cirruslib.transfer import get_s3_session
from cirruslib.utils import get_path

# envvars
CATALOG_BUCKET = os.getenv('CIRRUS_CATALOG_BUCKET', None)
PUBLISH_TOPIC_ARN = os.getenv('CIRRUS_PUBLISH_TOPIC_ARN', None)

# clients
statedb = StateDB()
snsclient = boto3.client('sns')
stepfunctions = boto3.client('stepfunctions')

# logging
logger = logging.getLogger(__name__)


class Catalog(dict):

    def __init__(self, *args, update=False, state_item=None, **kwargs):
        """Initialize a Catalog, verify required fields, and assign an ID

        Args:
            state_item (Dict, optional): Dictionary of entry in StateDB. Defaults to None.
        """
        super(Catalog, self).__init__(*args, **kwargs)

        # convert old functions field to tasks
        if 'functions' in self['process']:
            self['process']['tasks'] = self['process'].pop('functions')

        self.logger = get_task_logger(__name__, catalog=self)

        if update:
            self.update()

        # validate process block
        assert(self['type'] == 'FeatureCollection')
        assert('process' in self)
        assert('output_options' in self['process'])
        assert('workflow' in self['process'])
        assert('tasks' in self['process'])
        assert('workflow-' in self['id'])

        # TODO - validate with a JSON schema
        #if schema:
        #    pass
        # For now, just make check that there is at least one item
        assert(len(self['features']) > 0)
        for item in self['features']:
            if 'links' not in item:
                item['links'] = [] 

        # update collection IDs of member Items
        self.assign_collections()

        self.state_item = state_item

    @classmethod
    def from_payload(cls, payload: Dict, **kwargs) -> Catalog:
        """Parse a Cirrus payload and return a Catalog instance

        Args:
            payload (Dict): A payload from SNS, SQS, or containing an s3 URL to payload

        Returns:
            Catalog: A Catalog instance
        """
        if 'Records' in payload:
            records = [json.loads(r['body']) for r in payload['Records']]
            # there should be only one
            assert(len(records) == 1)
            if 'Message' in records[0]:
                # SNS
                cat = json.loads(records[0]['Message'])
            else:
                # SQS
                cat = records[0]
        elif 'url' in payload:
            cat = s3().read_json(payload['url'])
        elif 'Parameters' in payload and 'url' in payload['Parameters']:
            # this is Batch, get the output payload
            url = payload['Parameters']['url'].replace('.json', '_out.json')
            cat = s3().read_json(url)
        else:
            cat = payload
        return cls(cat, **kwargs)

    def update(self):
        if 'collections' in self['process']:
            # allow overriding of collections name 
            collections_str = self['process']['collections']
        else:
            # otherwise, get from items
            cols = sorted(list(set([i['collection'] for i in self['features'] if 'collection' in i])))
            input_collections = cols if len(cols) != 0 else 'none'
            collections_str = '/'.join(input_collections)
        
        items_str = '/'.join(sorted(list([i['id'] for i in self['features']])))
        if 'id' not in self:
            self['id'] = f"{collections_str}/workflow-{self['process']['workflow']}/{items_str}"

    # assign collections to Items given a mapping of Col ID: ID regex
    def assign_collections(self):
        """Assign new collections to all Items (features) in Catalog
            based on self['process']['output_options']['collections']
        """
        collections = self['process']['output_options'].get('collections', {})
        # loop through all Items in Catalog
        for item in self['features']:
            # loop through all provided output collections regexs
            for col in collections:
                regex = re.compile(collections[col])
                if regex.match(item['id']):
                    self.logger.debug(f"Setting collection to {col}")
                    item['collection'] = col

    def get_payload(self) -> Dict:
        """Get original payload for this Catalog

        Returns:
            Dict: Cirrus Input Catalog
        """
        payload = json.dumps(self)
        if CATALOG_BUCKET and len(payload.encode('utf-8')) > 30000:
            assert(CATALOG_BUCKET)
            url = f"s3://{CATALOG_BUCKET}/payloads/{uuid.uuid1()}.json"
            s3().upload_json(self, url)
            return {'url': url}
        else:
            return dict(self)

    # publish the items in this catalog
    def publish_to_s3(self, bucket, public=False) -> List:
        """Publish all Items to s3

        Args:
            bucket (str): Name of bucket to publish to
            public (bool, optional): Make published STAC Item public. Defaults to False.

        Returns:
            List: List of s3 URLs to published Items
        """
        opts = self['process'].get('output_options', {})
        s3urls = []
        for item in self['features']:
            # determine URL of data bucket to publish to- always do this
            url = os.path.join(get_path(item, opts.get('path_template')), f"{item['id']}.json")
            if url[0:5] != 's3://':
                url = f"s3://{bucket}/{url.lstrip('/')}"
            if public:
                url = s3.s3_to_https(url)

            # add canonical and self links (and remove existing self link if present)
            item['links'] = [l for l in item['links'] if l['rel'] not in ['self', 'canonical']]
            item['links'].insert(0, {
                'rel': 'canonical',
                'href': url,
                'type': 'application/json'
            })
            item['links'].insert(0, {
                'rel': 'self',
                'href': url,
                'type': 'application/json'
            })

            # get s3 session
            s3session = get_s3_session(s3url=url)

            # if existing item use created date
            now = datetime.now(timezone.utc).isoformat()
            created = None
            if s3session.exists(url):
                old_item = s3session.read_json(url)
                created = old_item['properties'].get('created', None)
            if created is None:
                created = now
            item['properties']['created'] = created
            item['properties']['updated'] = now

            # publish to bucket
            headers = opts.get('headers', {})
            
            extra = {'ContentType': 'application/json'}
            extra.update(headers)
            s3session.upload_json(item, url, public=public, extra=extra)
            s3urls.append(url)
            self.logger.info("Published to s3")

        return s3urls

    @classmethod
    def sns_attributes(self, item) -> Dict:
        """Create attributes from Item for publishing to SNS

        Args:
            item (Dict): A STAC Item

        Returns:
            Dict: Attributes for SNS publishing
        """
        attr = {
            'collection': {
                'DataType': 'String',
                'StringValue': item['collection']
            },
            'datetime': {
                'DataType': 'String',
                'StringValue': item['properties']['datetime']
            },
            'bbox.ll_lon': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][0])
            },
            'bbox.ll_lat': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][1])
            },
            'bbox.ur_lon': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][2])
            },
            'bbox.ur_lat': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][3])
            }     
        }
        if 'eo:cloud_cover' in item['properties']:
            attr['cloud_cover'] = {
                'DataType': 'Number',
                'StringValue': str(item['properties']['eo:cloud_cover'])
            }
        if item['properties']['created'] != item['properties']['updated']:
            attr['status'] = {
                'DataType': 'String',
                'StringValue': 'updated'
            }
        else:
            attr['status'] = {
                'DataType': 'String',
                'StringValue': 'created'
            }
        return attr

    def publish_to_sns(self, topic_arn=PUBLISH_TOPIC_ARN):
        """Publish this catalog to SNS

        Args:
            topic_arn (str, optional): ARN of SNS Topic. Defaults to PUBLISH_TOPIC_ARN.
        """
        for item in self['features']:
            response = snsclient.publish(TopicArn=topic_arn, Message=json.dumps(item),
                                        MessageAttributes=self.sns_attributes(item))         
            self.logger.debug(f"Published item to {topic_arn}")

    def process(self) -> str:
        """Add this Catalog to Cirrus and start workflow

        Returns:
            str: Catalog ID
        """
        assert(CATALOG_BUCKET)

        arn = os.getenv('BASE_WORKFLOW_ARN') + self['process']['workflow']

        # start workflow
        try:
            # add input catalog to s3
            url = f"s3://{CATALOG_BUCKET}/{self['id']}/input.json"
            s3().upload_json(self, url)

            # create DynamoDB record - this overwrites existing states other than PROCESSING
            resp = statedb.claim_processing(self['id'])

            # invoke step function
            self.logger.debug(f"Running Step Function {arn}")
            exe_response = stepfunctions.start_execution(stateMachineArn=arn, input=json.dumps(self.get_payload()))

            # add execution to DynamoDB record
            resp = statedb.set_processing(self['id'], exe_response['executionArn'])

            return self['id']
        except statedb.db.meta.client.exceptions.ConditionalCheckFailedException:
            msg = f"Already in PROCESSING state"
            self.logger.warning(msg)
            return None
        except Exception as err:
            msg = f"failed starting workflow ({err})"
            self.logger.error(msg, exc_info=True)
            statedb.set_failed(self['id'], msg)
            raise err


class Catalogs(object):

    def __init__(self, catalogs, state_items=None):
        self.catalogs = catalogs
        if state_items:
            assert(len(state_items) == len(self.catalogs))
        self.state_items = state_items

    def __getitem__(self, index):
        return self.catalogs[index]

    @property
    def catids(self) -> List[str]:
        """Return list of catalog IDs

        Returns:
            List[str]: List of Catalog IDs
        """
        return [c['id'] for c in self.catalogs]

    @classmethod
    def from_catids(cls, catids: List[str], **kwargs) -> Catalogs:
        """Create Catalogs from list of Catalog IDs

        Args:
            catids (List[str]): List of catalog IDs

        Returns:
            Catalogs: A Catalogs instance
        """
        items = [statedb.dbitem_to_item(statedb.get_dbitem(catid)) for catid in catids]
        catalogs = []
        for item in items:
            cat = Catalog(s3().read_json(item['catalog']))
            catalogs.append(cat)
        logger.debug(f"Retrieved {len(catalogs)} from state db")
        return cls(catalogs, state_items=items)

    """
    @classmethod
    def from_statedb_paged(cls, collections, state, since: str=None, index: str='input_state', limit=None):
        catalogs = []
        # get first page
        resp = statedb.get_items_page(collections, state, since, index)
        for it in resp['items']:
            cat = Catalog(s3().read_json(it['input_catalog']))
            catalogs.append(cat)
        self.logger.debug(f"Retrieved {len(catalogs)} from state db")
        yield cls(catalogs, state_items=resp['items'])
        catalogs = []
        while 'nextkey' in resp:
            resp = statedb.get_items_page(collections, state, since, index, nextkey=resp['nextkey'])
            for it in resp['items']:
                cat = Catalog(s3().read_json(it['input_catalog']))
                catalogs.append(cat)
            self.logger.debug(f"Retrieved {len(catalogs)} from state db")
            yield cls(catalogs, state_items=resp['items'])
    """

    @classmethod
    def from_statedb(cls, collections, state, since: str=None, index: str='input_state', limit=None) -> Catalogs:
        """Create Catalogs object from set of StateDB Items

        Args:
            collections (str): String of collections (input or output depending on `index`)
            state (str): The state (QUEUED, PROCESSING, COMPLETED, FAILED, INVALID) of StateDB Items to get
            since (str, optional): Get Items since this duration ago (e.g., 10m, 8h, 1w). Defaults to None.
            index (str, optional): 'input_state' or 'output_state' Defaults to 'input_state'.
            limit ([type], optional): Max number of Items to return. Defaults to None.

        Returns:
            Catalogs: Catalogs instance
        """
        catalogs = []
        items = statedb.get_items(collections, state, since, index, limit=limit)
        logger.debug(f"Retrieved {len(items)} total items from statedb")
        for item in items:
            cat = Catalog(s3().read_json(item['catalog']))
            catalogs.append(cat)
        logger.debug(f"Retrieved {len(catalogs)} input catalogs")
        return cls(catalogs, state_items=items)

    def get_states(self):
        if self.state_items is None:
            items = [statedb.dbitem_to_item(i) for i in statedb.get_dbitems(self.catids)]
            self.state_items = items
        states = {c['catid']: c['state'] for c in self.state_items}
        return states

    def process(self, replace=False):
        """Create Item in Cirrus State DB for each Catalog and add to processing queue

        Args:
            catalog (Dict): A Cirrus Input Catalog
        """
        catids = []
        # check existing states            
        states = self.get_states()
        for cat in self.catalogs:
            _replace = replace or cat['process'].get('replace', False)
            # check existing state for Item, if any
            state = states.get(cat['id'], '')
            # don't try and process these - if they are stuck they should be removed from db
            #if state in ['QUEUED', 'PROCESSING']:
            #    logger.info(f"Skipping {cat['id']}, in {state} state")
            #    continue
            if state in ['FAILED', ''] or _replace:
                catid = cat.process()
                if catid is None:
                    catids.append(catid)
            else:
                logger.info(f"Skipping, input already in {state} state")
                continue

        return catids



import boto3
import json
import logging
import os

from boto3utils import s3
from typing import Dict, Optional, List

from pystac import STAC_IO, Catalog, CatalogType, Collection, Link

logger = logging.getLogger(__name__)

# envvars
DATA_BUCKET = os.getenv('CIRRUS_DATA_BUCKET', None)
STAC_VERSION = os.getenv('CIRRUS_STAC_VERSION', '1.0.0-beta.2')
DESCRIPTION = os.getenv('CIRRUS_STAC_DESCRIPTION', 'Cirrus STAC')
AWS_REGION = os.getenv('AWS_REGION')
PUBLISH_TOPIC = os.getenv('CIRRUS_PUBLISH_TOPIC_ARN', None)
PUBLIC_CATALOG = os.getenv('CIRRUS_PUBLIC_CATALOG', False)
if isinstance(PUBLIC_CATALOG, str):
    PUBLIC_CATALOG = True if PUBLIC_CATALOG.lower() == 'true' else False

# root catalog
ROOT_URL = f"s3://{DATA_BUCKET}"
if PUBLIC_CATALOG:
    ROOT_URL = s3.s3_to_https(ROOT_URL, region=AWS_REGION)

# boto3 clients
snsclient = boto3.client('sns')


def s3stac_read(uri):
    if uri.startswith('http'):
        uri = s3.https_to_s3(uri)
    return json.dumps(s3().read_json(uri))

def s3stac_write(uri, txt):
    extra = {
        'ContentType': 'application/json'
    }
    if uri.startswith('http'):
        uri = s3.https_to_s3(uri)
    s3().upload_json(json.loads(txt), uri, extra=extra, public=PUBLIC_CATALOG)

STAC_IO.read_text_method = s3stac_read
STAC_IO.write_text_method = s3stac_write


def get_root_catalog():
    """Get Cirrus root catalog from s3

    Returns:
        Dict: STAC root catalog
    """
    caturl = f"{ROOT_URL}/catalog.json"
    if s3().exists(caturl):
        cat = Catalog.from_file(caturl)
    else:
        catid = DATA_BUCKET.split('-data-')[0]
        cat = Catalog(id=catid, description=DESCRIPTION)
        cat.normalize_and_save(ROOT_URL, CatalogType.ABSOLUTE_PUBLISHED)
    logger.debug(f"Fetched {cat.describe()}")
    return cat


# add this collection to Cirrus catalog
def add_collections(collections, publish=True):

    for collection in collections:
        collection.remove_links('child')
        link = Link('copied_from', collection)
        collection.add_link(link, collection.get_self_href())
        ROOT_CATALOG.add_child(collection)
        if publish:
            child_json = json.dumps(collection.to_dict())
            logger.debug(f"Publishing {collection.id}: {child_json}")
            response = snsclient.publish(TopicArn=PUBLISH_TOPIC, Message=child_json)
            logger.debug(f"SNS Publish response: {json.dumps(response)}")
 
    ROOT_CATALOG.normalize_and_save(ROOT_URL, CatalogType.ABSOLUTE_PUBLISHED)
    return ROOT_CATALOG


ROOT_CATALOG = get_root_catalog()


import boto3
import json
import logging
import requests
import uuid

from boto3utils import s3
from dateutil.parser import parse as dateparse
from os import getenv
from string import Formatter, Template
from typing import Dict, Optional, List
from collections.abc import Mapping

logger = logging.getLogger(__name__)

batch_client = boto3.client('batch')


def submit_batch_job(payload, arn, queue='basic-ondemand', definition='geolambda-as-batch', name=None):
    # envvars
    STACK_PREFIX = getenv('CIRRUS_STACK')
    CATALOG_BUCKET = getenv('CIRRUS_CATALOG_BUCKET')

    if name is None:
        name = arn.split(':')[-1]

    # upload payload to s3
    url = f"s3://{CATALOG_BUCKET}/batch/{uuid.uuid1()}.json"
    s3().upload_json(payload, url)
    kwargs = {
        'jobName': name,
        'jobQueue': f"{STACK_PREFIX}-{queue}",
        'jobDefinition': f"{STACK_PREFIX}-{definition}",
        'parameters': {
            'lambda_function': arn,
            'url': url
        },
        'containerOverrides': {
            'vcpus': 1,
            'memory': 512,
        }
    }
    logger.debug(f"Submitted batch job with payload {url}")
    response = batch_client.submit_job(**kwargs)
    logger.debug(f"Batch response: {response}")


def get_path(item: Dict, template: str='${collection}/${id}') -> str:
    """Get path name based on STAC Item and template string

    Args:
        item (Dict): A STAC Item.
        template (str, optional): Path template using variables referencing Item fields. Defaults to '${collection}/${id}'.

    Returns:
        [str]: A path name
    """
    _template = template.replace(':', '__colon__')
    subs = {}
    for key in [i[1] for i in Formatter().parse(_template.rstrip('/')) if i[1] is not None]:
        # collection
        if key == 'collection':
            subs[key] = item['collection']
        # ID
        elif key == 'id':
            subs[key] = item['id']
        # derived from date
        elif key in ['year', 'month', 'day']:
            dt = dateparse(item['properties']['datetime'])
            vals = {'year': dt.year, 'month': dt.month, 'day': dt.day}
            subs[key] = vals[key]
        # Item property
        else:
            subs[key] = item['properties'][key.replace('__colon__', ':')]
    return Template(_template).substitute(**subs).replace('__colon__', ':')


# from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9#gistcomment-2622319
def dict_merge(dct, merge_dct, add_keys=True):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    This version will return a copy of the dictionary and leave the original
    arguments untouched.
    The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dict`` but not ``dct`` should be included in the
    new dict.
    Args:
        dct (dict) onto which the merge is executed
        merge_dct (dict): dct merged into dct
        add_keys (bool): whether to add new keys
    Returns:
        dict: updated dict
    """
    dct = dct.copy()
    if not add_keys:
        merge_dct = {
            k: merge_dct[k]
            for k in set(dct).intersection(set(merge_dct))
        }

    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], Mapping)):
            dct[k] = dict_merge(dct[k], merge_dct[k], add_keys=add_keys)
        else:
            dct[k] = merge_dct[k]

    return dct


import boto3
import json
import logging
import requests

from boto3utils import s3, secrets
from botocore.exceptions import ClientError
from copy import deepcopy
from dateutil.parser import parse as dateparse
from os import getenv, path as op
from cirruslib.utils import get_path

from typing import Dict, Optional, List

logger = logging.getLogger(__name__)

# get data bucket to upload to
DATA_BUCKET = getenv('CIRRUS_DATA_BUCKET')

## global dictionary of sessions per bucket
s3_sessions = {}


def get_s3_session(bucket: str=None, s3url: str=None, **kwargs) -> s3:
    """Get boto3-utils s3 class for interacting with an s3 bucket. A secret will be looked for with the name
    `cirrus-creds-<bucket-name>`. If no secret is found the default session will be used

    Args:
        bucket (str, optional): Bucket name to access. Defaults to None.
        url (str, optional): The s3 URL to access. Defaults to None.

    Returns:
        s3: A boto3-utils s3 class
    """
    if s3url:
        parts = s3.urlparse(s3url)
        bucket = parts['bucket']

    if bucket and bucket in s3_sessions:
        return s3_sessions[bucket]
    # otherwise, create new session for this bucket
    creds = deepcopy(kwargs)
    
    try:
        # get credentials from AWS secret
        secret_name = f"cirrus-creds-{bucket}"
        _creds = secrets.get_secret(secret_name)
        creds.update(_creds)
        logger.debug(f"Using credentials for bucket {bucket}: {json.dumps(creds)}")
    except ClientError:
        # using default credentials
        pass
        

    requester_pays = creds.pop('requester_pays', False)
    session = boto3.Session(**creds)
    s3_sessions[bucket] = s3(session, requester_pays=requester_pays)
    return s3_sessions[bucket]


def download_from_http(url: str, path: str='') -> str:
    """ Download a file over http and save to path

    Args:
        url (str): A URL to download
        path (str, optional): A local path name to save file. Defaults to '' (current directory)

    Returns:
        str: Local filename of saved file. Basename is the same as the URL basename
    """
    filename = op.join(path, op.basename(url))
    resp = requests.get(url, stream=True)
    with open(filename, 'wb') as f:
        for chunk in resp.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
    return filename


def download_item_assets(item: Dict, path: str='', assets: Optional[List[str]]=None) -> Dict:
    """Download STAC Item assets to local filesystem

    Args:
        item (Dict): A STAC Item dictionary
        path (str, optional): Path to download files to. Defaults to current directory
        assets (Optional[List[str]], optional): List of asset keys to download. Defaults to all assets
        s3_session (s3, optional): boto3-utils s3 object for s3 interactions. Defaults to None

    Returns:
        Dict: A new STAC Item with downloaded assets pointing to newly downloaded files
    """

    # if assets not provided, download all assets
    assets = assets if assets is not None else item['assets'].keys()

    _item = deepcopy(item)

    for a in assets:
        # download each asset
        url = item['assets'][a]['href']
        logger.debug(f"Downloading {url}")

        # http URL to s3 source
        if 'amazonaws.com' in url:
            url = s3.https_to_s3(url)

        filename = None
        # s3 source
        if url.startswith('s3://'):
            parts = s3.urlparse(url)
            s3_session = get_s3_session(parts['bucket'])
            filename = s3_session.download(url, path=path)
        # general http source
        elif url.startswith('http'):
            filename = download_from_http(url, path=path)
        else:
            logger.error(f"Unknown protocol for {url}")

        # if downloaded update href in Item
        if filename:
            _item['assets'][a]['href'] = op.abspath(filename)
    return _item


def upload_item_assets(item: Dict, assets: List[str]=None, public_assets: List[str]=[], 
                       path_template: str='${collection}/${id}', s3_urls: bool=False,
                       headers: Dict={}, s3_session: s3=None, **kwargs) -> Dict:
    """Upload Item assets to s3 bucket

    Args:
        item (Dict): STAC Item
        assets (List[str], optional): List of asset keys to upload. Defaults to None.
        public_assets (List[str], optional): List of assets keys that should be public. Defaults to [].
        path_template (str, optional): Path string template. Defaults to '${collection}/${id}'.
        s3_urls (bool, optional): Return s3 URLs instead of http URLs. Defaults to False.
        headers (Dict, optional): Dictionary of headers to set on uploaded assets. Defaults to {}.
        s3_session (s3, optional): boto3-utils s3 object for s3 interactions. Defaults to None

    Returns:
        Dict: A new STAC Item with uploaded assets pointing to newly uploaded file URLs
    """
    # if assets not provided, upload all assets
    _assets = assets if assets is not None else item['assets'].keys()

    # determine which assets should be public
    if type(public_assets) is str and public_assets == 'ALL':
        public_assets = item['assets'].keys()

    # deepcopy of item
    _item = deepcopy(item)

    for key in [a for a in _assets if a in item['assets'].keys()]:
        asset = item['assets'][key]
        filename = asset['href']
        if not op.exists(filename):
            logger.warning(f"Cannot upload {filename}: does not exist")
            continue
        public = True if key in public_assets else False
        _headers = {}
        if 'type' in asset:
            _headers['ContentType'] = asset['type']
        _headers.update(headers)
        # output URL
        url = get_path(item, op.join(path_template, op.basename(filename)))
        # if output URL is relative, put it in the default data bucket
        if url[0:5] != 's3://':
            url = f"s3://{DATA_BUCKET}/{url}"
        parts = s3.urlparse(url)
        s3_session = get_s3_session(parts['bucket'])

        # upload
        logger.debug(f"Uploading {filename} to {url}")
        url_out = s3_session.upload(filename, url, public=public, extra=_headers, http_url=not s3_urls)
        _item['assets'][key]['href'] = url_out
    return _item


